import os, sys
sys.path.append(os.path.abspath('../tool'))

import math
import behavior
import effect
import predicate
import logging

import defensive
import obj

test = defensive.Testing()
log = defensive.Logging('bubble')


@obj.immutable
class Bubble(object):
    def __init__(self,name,\
                 (x,y),r,\
                 (dirx,diry),
                 (vx,vy),\
                 child,\
                 link,\
                 predicate,\
                 behavior,\
                 tag):
        self.name = name
        self.position= (x,y)
        self.r = r
        self.direction = (dirx,diry)
        self._forceVector = (vx,vy)
        self._child = tuple(child)
        self.link = tuple(link)
        self.predicate = predicate
        self.behavior = tuple(behavior)
        self.tag = tuple(tag)
	    
    def __repr__(self):
        return 'Bubble['+self.name+']'
    
    def __str__(self):
        return 'Bubble('+self.name+',at'+self.position+',look'+self.direction+',go'+self._forceVector+')'

    @test.eq((1,1))
    @log.debug
    def getNextPos(self):
        """
        return predicted x,y position
        """
        x,y = self.position
        vx,vy = self.getForceVector()
        return (x+vx,y+vy)

    @log.debug  
    def getChild(self):
        """
        return translated child
        """
        for child in self._child:
            x,y = self.position
            yield child.translate(x, y)

    @log.debug
    def getRadian(self):
        """
        return direction angle in radian
        """
        dirx,diry = self.direction
        return math.atan2(dirx,diry)

    @log.debug
    def getDelta(self,other):
        """
        return delta of respective (x,y) position
        """
        ax,ay = self.position
        bx,by = other.position
        deltaX = ax - bx
        deltaY = ay - by
        return deltaX,deltaY
    
    @log.debug
    def getDistSQ(self,other):
        """
        return square of distance between two bubble
        """
        deltaX,deltaY = self.getDelta(other)
        return deltaX**2 + deltaY**2
    
    @log.debug
    def intersect(self,other):
        """
        return true if two bubble intersect
        """
        minDist = self.r + other.r
        minDistSQ = minDist**2
        return minDistSQ >= self.getDistSQ(other)
    
    @log.debug
    def inside(self,other):
        """
        return true if self is completely inside the other
        """
        minDist = other.r - self.r
        minDistSQ = minDist**2
        return minDistSQ >= self.getDistSQ(other)

    @log.debug
    def getForceVector(self):
        """
        return forcevector and pseudo normalise it
        """
        vx,vy = self._forceVector
        if vx > 1:
            vx = 1
        if vx < -1:
            vx = -1
        if vy > 1:
            vy = 1
        if vy < -1:
            vy = -1
        return vx,vy   

    @log.debug
    @obj.immutator
    def stop(copy):
        """
        return modified bubble for forceVector as (0,0)
        """
        copy._forceVector = (0,0)
        return copy
    
    @log.debug
    @obj.immutator
    def translate(copy,x,y):
        """
        return modified bubble with new position
        """
        ax,ay = copy.position
        copy.position=(ax+x,ay+y)
        return copy
    
    @log.debug
    @obj.immutator
    def rotate(copy,rad):
        """
        return modified bubble with new direction
        """
        s = math.sin(rad)
        c = math.cos(rad)
        x,y = copy.direction
        dirx = x*c - y*s
        diry = x*s +y*c
        dist = math.hypot(dirx,diry)
        dirx /= dist
        diry /= dist
        copy.direction = (dirx,diry)
        return copy
       
    @log.debug   
    @obj.immutator            
    def setDirection(copy,dirx,diry):
        """
        return modified bubble with new direction
        """
        copy.direction = (dirx,diry)
        return copy
    
    @log.debug
    @obj.immutator
    def push(copy,x,y):
        """
        return modified bubble with updated forceVector
        """
        vx,vy = copy.getForceVector()
        vx = vx+x
        vy = vy+y
        copy._forceVector = (vx,vy)
        return copy

    @log.debug
    @obj.immutator
    def restrain(copy,scale):
        """
        return modified bubble with updated forceVector
        """
        vx,vy = copy.getForceVector()
        vx = vx*scale
        vy = vy*scale
        copy._forceVector = (vx,vy)
        return copy

    @log.debug
    @obj.immutator
    def giggle(copy):
        """
        return modified bubble with applied force vector on position
        """
        x,y = copy.position
        vx, vy = copy.getForceVector()
        copy.position = (x+vx,y+vy)
        return copy

    @log.debug
    @obj.immutator
    def updateChild(copy,child):
        """
        return modified bubble with new child list
        """
        copy._child = tuple(child)
        return copy
    
    @log.debug
    @obj.immutator
    def updateLink(copy,newlink):
        """
        return modified bubble with new link list
        """
        copy.link = tuple(newlink)
        return copy

    @log.debug
    @obj.immutator
    def addTag(copy,newTag):
        """
        return modified bubble with one tag added
        """
        copy.tag = copy.tag + (newTag,)
        return copy
      
    @log.debug  
    @obj.immutator
    def rmTag(copy,rmTag):
        """
        return modified bubble with one tag removed
        """
        copy.tag = tuple(list(copy.tag).remove(rmTag))
        return copy                   
      
    @log.debug        
    def recursivePredicate(self,bubble,cnt=1):
        """
        return the child the bubble is in
        """
        if self.predicate(bubble,self):
            for child in self.getChild():
                if child.predicate(bubble,child):
                    return child.recursivePredicate(bubble,cnt+1)
        return self

    #@logger.debug
    @obj.immutator
    def against(copy,context,cnt=1):
        """
        return modified bubble after all behaviors have been executed
        """
        inbubble = context.recursivePredicate(copy)
        newSelf = copy
        for behavior in newSelf.behavior:
            newSelf = behavior(newSelf,inbubble)
        newLink=[]
        for link in newSelf.link:
            newLink.append(link.against(context,cnt+1))
        return newSelf.updateLink(newLink)

#some function
@log.debug
def bubbleFromStruct(db,effectDict,behaviorDict,predicateDict):
    """
    return one bubble list from given json data
    """
    cluster = []
    for name in db.keys():
        cluster.append(Bubble(name,\
                              (db[name]['x'],db[name]['y']),\
                               db[name]['r'],\
                              (db[name]['dirx'],db[name]['diry']),\
                              (db[name]['vx'],db[name]['vy']),\
                               bubbleFromStruct(db[name]['child'],effectDict,behaviorDict,predicateDict),\
                               bubbleFromStruct(db[name]['link'],effectDict,behaviorDict,predicateDict),\
                               predicateDict[db[name]['predicate']],\
                               (behaviorDict[behavior] for behavior in db[name]['behavior']),\
                               db[name]['tag']))
    return cluster

if __name__ == '__main__':
    import logging
    import logging.handlers
    ch = logging.handlers.SocketHandler('localhost',
                                                   logging.handlers.DEFAULT_TCP_LOGGING_PORT)

    #ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    log.logger.addHandler(ch)
    
    bubble = Bubble('test',\
                 (0,0),1,\
                 (0,1),
                 (1,1),\
                 (),\
                 (),\
                 lambda a,b:True,\
                 (),\
                 ())
    test.runTest(bubble)
